#include "stdafx.h"
#include "CCaptionPainter.h"
#include <process.h>
#include <vector>
#include "CommRoutine.h"

#define DEFAULT_OSD_COUNT		30
#define DEFAULT_DELAY_TIME		1300	// 1.3 sec

COSDManager::COSDManager(HWND hParent, int countOSD)
{
	this->m_hParentView = hParent;

	InitializeCriticalSection(&this->m_csList);

	// prepare OSDs.
	for(int i = 0; i < countOSD; i++) {	
		CPseudoOSD * pOSD = new CPseudoOSD();
		pOSD->Initialize((HINSTANCE)GetModuleHandle(NULL));
		pOSD->Create(hParent);
		pOSD->Hide();

		this->m_freeOSDList.push_back(pOSD);
	}
}

COSDManager::~COSDManager()
{
	OSD_LIST tmpList;
	
	EnterCriticalSection(&this->m_csList);

	OSD_LIST::iterator it = this->m_freeOSDList.begin();

	for(; it != this->m_freeOSDList.end(); it++)
		tmpList.push_back(*it);

	m_freeOSDList.clear();

	for(it = m_inuseOSDList.begin(); it != m_inuseOSDList.end(); it++)
		tmpList.push_back(*it);

	m_inuseOSDList.clear();

	LeaveCriticalSection(&this->m_csList);

	for(it = tmpList.begin(); it != tmpList.end(); it++)
		delete (*it);

	DeleteCriticalSection(&this->m_csList);
}


CPseudoOSD * COSDManager::GetOSD()
{
	CPseudoOSD * pOSD = NULL;

	EnterCriticalSection(&this->m_csList);

	if(this->m_freeOSDList.size() > 0) {
		pOSD = this->m_freeOSDList.front();
		this->m_freeOSDList.pop_front();
	}

	LeaveCriticalSection(&this->m_csList);

	return pOSD;
}

void COSDManager::MarkAsInUse(CPseudoOSD * pOSD)
{
	EnterCriticalSection(&this->m_csList);

	this->m_inuseOSDList.push_back(pOSD);

	LeaveCriticalSection(&this->m_csList);
}

void COSDManager::OSDAllHide()
{
	CPseudoOSD * pOSD = NULL;
	OSD_LIST tmpList;

	EnterCriticalSection(&this->m_csList);

	OSD_LIST::iterator it = this->m_inuseOSDList.begin();

	for(; it != this->m_inuseOSDList.end(); it++) {
		pOSD = (*it);
		tmpList.push_back(pOSD);
	}

	this->m_inuseOSDList.clear();

	LeaveCriticalSection(&this->m_csList);

	for(it = tmpList.begin(); it != tmpList.end(); it++) {
		pOSD = (*it);
		pOSD->Hide();

		EnterCriticalSection(&this->m_csList);
		this->m_freeOSDList.push_back(pOSD);
		LeaveCriticalSection(&this->m_csList);
	}
}

SIZE COSDManager::OSDShow(OSD_DISPLAY_PARM parm)
{
	CPseudoOSD * pOSD = NULL;

	SIZE sz;
	sz.cx = 0;
	sz.cy = 0;

	pOSD = GetOSD();

	if(!pOSD)
		return sz;

	pOSD->SetTextColor(parm.color);
	pOSD->SetTextHeight(parm.textHeight);
	pOSD->SetText(parm.text);
	pOSD->CalcTextSize(&sz);

	pOSD->SetPosition(
		parm.left,
		parm.top,
		sz.cx,
		sz.cy
		);

	pOSD->Show();

	MarkAsInUse(pOSD);

	return sz;
}

CCaptionPainter::CCaptionPainter(HWND hViewer)
{
	this->m_hViewer = hViewer;
	this->m_delayTime = DEFAULT_DELAY_TIME;
	this->m_pOSDManager = new COSDManager(hViewer, DEFAULT_OSD_COUNT);

	m_hNewCaptionEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	m_hListMutex = CreateMutex(NULL, FALSE, NULL);
	m_hTerminateEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

	m_hDisplayThreadHandle = (HANDLE)_beginthreadex(
		NULL,
		0,
		CCaptionPainter::CaptionDisplayProc,
		this,
		0,
		NULL
		);
}

CCaptionPainter::~CCaptionPainter()
{
	SetEvent(m_hTerminateEvent);

	if(m_hDisplayThreadHandle) {
		WaitForSingleObject(m_hDisplayThreadHandle, INFINITE);
		DbgString(_T("Thread Terminated.\n"));
	}

	delete this->m_pOSDManager;
}

bool CCaptionPainter::Initialize()
{
	if(NO_ERR == this->m_captionUtil.Initialize())
		return TRUE;

	return FALSE;
}

void CCaptionPainter::Clear()
{
	WaitForSingleObject(this->m_hListMutex, INFINITE);

	this->m_showCaptionList.clear();

	//
	// insert [CS] command into list.
	//
	CAPTION_DISPLAY clear;

	clear.data.wSWFMode = 0;
	clear.data.bClear = TRUE;
	clear.data.dwWaitTime = 0;
	clear.showTime.QuadPart = 0;

	this->m_showCaptionList.push_back(clear);

	SetEvent(this->m_hNewCaptionEvent);

	ReleaseMutex(this->m_hListMutex);
}

void CCaptionPainter::AddCaptionPacket(unsigned char *pPacket)
{
	int ret = this->m_captionUtil.AddTSPacket(pPacket);

	if(ret == CHANGE_VERSION) {
		DbgString(DBG_PREFIX _T("CaptionDLL CHANGE_VERSION\n"));
		std::vector<LANG_TAG_INFO> tagInfoList;
		ret = this->m_captionUtil.GetTagInfo(&tagInfoList);

		std::vector<LANG_TAG_INFO>::iterator it = tagInfoList.begin();
		for(;it!= tagInfoList.end(); it++) {
			WCHAR pwszBuff[32] = {0};
			MultiByteToWideChar( 932, 0, it->szISOLangCode, -1, pwszBuff, 32);
			int iData = (int)it->ucLangTag;

			DbgString(DBG_PREFIX _T("\t TagInfo: %s, LangTag %d\n"), pwszBuff, iData);
		}
	}
	else if(ret == NO_ERR_CAPTION) {
		
		std::vector<CAPTION_DATA> Captions;
		CAPTION_DISPLAY show;
		LARGE_INTEGER curTime;

		// hardcode...
		int index = 0;
		ret = this->m_captionUtil.GetCaptionData(index, &Captions);

		GetSystemTimeAsFileTime((FILETIME *)&curTime);

		WaitForSingleObject(this->m_hListMutex, INFINITE);

		std::vector<CAPTION_DATA>::iterator it = Captions.begin();
		for(;it != Captions.end(); it++) {
			CAPTION_DISPLAY show;

			show.data = *it;
			show.showTime.QuadPart = curTime.QuadPart + (this->m_delayTime + show.data.dwWaitTime) * 10000;

			this->m_showCaptionList.push_back(show);
		}

		SetEvent(this->m_hNewCaptionEvent);

		ReleaseMutex(this->m_hListMutex);
	}
	else if(ret == ERR_FALSE) {
		DbgString(DBG_PREFIX _T("CaptionDLL return FALSE\n"));
	}
	else if(ret == NO_ERR_TAG_INFO) {
		DbgString(DBG_PREFIX _T("CaptionDLL return NO_ERR_TAG_INFO\n"));
		std::vector<LANG_TAG_INFO> tagInfoList;
		this->m_captionUtil.GetTagInfo(&tagInfoList);

		std::vector<LANG_TAG_INFO>::iterator it = tagInfoList.begin();
		for(;it!= tagInfoList.end(); it++) {
			WCHAR pwszBuff[32] = {0};
			MultiByteToWideChar( 932, 0, it->szISOLangCode, -1, pwszBuff, 32);
			int iData = (int)it->ucLangTag;

			DbgString(DBG_PREFIX _T("\t TagInfo: %s, LangTag %d\n"), pwszBuff, iData);
		}
	}
	else if(ret != ERR_NEED_NEXT_PACKET) {
		DbgString(DBG_PREFIX _T("CaptionDLL return %d\n"), ret);
	}
	else if( ret == ERR_NEED_NEXT_PACKET) {
		DbgString(DBG_PREFIX _T("CaptionDLL need next packet\n"));
	}
}

void CCaptionPainter::SetDelayTime(unsigned int msec)
{
	this->m_delayTime = msec;
}

void CCaptionPainter::ShowCaptionData(CAPTION_DATA data)
{
	if(data.bClear) {
		this->m_pOSDManager->OSDAllHide();
		return;
	}

	RECT videoRect;
	GetWindowRect(this->m_hViewer, &videoRect);

	int videoWidth = videoRect.right - videoRect.left;
	int videoHight = videoRect.bottom - videoRect.top;
	float ratioY = 1;
	float ratioX = 1;

	int iPoX = data.wPosX;
	int iPoY = data.wPosY;

	switch(data.wSWFMode) {
		case 7:
			{
				// 960x540
				ratioY = ((float)videoHight / (float)540);
				ratioX = ((float)videoWidth / (float)960);

				iPoX = (int)((float)(data.wClientX + data.wPosX) * ratioX);
				iPoY = (int)((float)(data.wClientY + data.wPosY -50) * ratioY);
			}
			break;

		case 0:
			{
				ratioY = ((float)videoHight / (float)540);
				ratioX = ((float)videoWidth / (float)960);

				iPoY = data.wPosY + data.wClientY;
				iPoX = data.wPosX + data.wClientX;

				iPoY = (int)((float)iPoY * ratioY);
				iPoX = (int)((float)iPoX * ratioX);
			}
			break;

		default:
			{
				iPoY = data.wPosY + data.wClientY;
				iPoX = data.wPosX + data.wClientX;
			}
			break;
	}

	std::vector<CAPTION_CHAR_DATA>::iterator it2 = data.CharList.begin();

	for(;it2 != data.CharList.end(); it2++) {

		int iFontH = (int)((float)it2->wCharH * ratioY);
		int iFontW = (int)((float)it2->wCharW * ratioY);

		if( it2->emCharSizeMode == STR_SMALL ){
			iFontH = iFontH/2;

			if(data.wSWFMode == 0)
				iPoY += (int)((float)30 * ratioY);

			else if(data.wSWFMode == 7)
				iPoY += (int)((float)25 * ratioY);

		}else if( it2->emCharSizeMode == STR_MEDIUM ){
			// Sp -> p
			it2->strDecode = GetHalfChar(it2->strDecode);
		}else if( it2->emCharSizeMode == STR_HIGH_W ){
			//				iFontW = iFontW*2;
		}else if( it2->emCharSizeMode == STR_WIDTH_W ){
			iFontH = iFontH*2;
		}else if( it2->emCharSizeMode == STR_W ){
			//				iFontW = iFontW*2;
			iFontH = iFontH*2;
		}

		WCHAR str[1024] = {0};
		MultiByteToWideChar(932, 0, it2->strDecode.c_str(), -1, str, 1024);

		DbgString(DBG_PREFIX _T("Font:%d PosX: %d PoxY: %d Cap:%ws \n"),iFontH, iPoX, iPoY, str);

		COSDManager::OSD_DISPLAY_PARM OSDParm;

		OSDParm.color = RGB(it2->stCharColor.ucR, it2->stCharColor.ucG, it2->stCharColor.ucB);
		OSDParm.text = str;
		OSDParm.textHeight = iFontH;
		OSDParm.left = iPoX;
		OSDParm.top = iPoY;

		SIZE sz = this->m_pOSDManager->OSDShow(OSDParm);

		iPoX += sz.cx;
	}
}

unsigned int WINAPI CCaptionPainter::CaptionDisplayProc(LPVOID lpParameter)
{
	CCaptionPainter *pThis = (CCaptionPainter *)lpParameter;
	HANDLE hs[] = {pThis->m_hTerminateEvent, pThis->m_hNewCaptionEvent};

	while(TRUE) {

		if(WAIT_OBJECT_0 == WaitForMultipleObjects(2, hs, FALSE, INFINITE)) {
			DbgString(_T("Terminate.\n"));
			break;
		}

		WaitForSingleObject(pThis->m_hListMutex, INFINITE);

		CAPTION_DISPLAY_LIST::iterator showIt = pThis->m_showCaptionList.begin();

		for(; showIt != pThis->m_showCaptionList.end();) {
			LARGE_INTEGER curTime;
			GetSystemTimeAsFileTime((FILETIME *)&curTime);

			if(showIt->showTime.QuadPart > curTime.QuadPart) {
				ReleaseMutex(pThis->m_hListMutex);

				Sleep((showIt->showTime.QuadPart - curTime.QuadPart) / 10000);
				break;
			}

			pThis->ShowCaptionData(showIt->data);

			CAPTION_DISPLAY_LIST::iterator _it = showIt;
			showIt++;
			pThis->m_showCaptionList.erase(_it);
		}

		if(pThis->m_showCaptionList.size() == 0)
			ResetEvent(pThis->m_hNewCaptionEvent);

		ReleaseMutex(pThis->m_hListMutex);

	};


	return 0;
}
