#include "stdafx.h"
#include <windows.h>
#include <process.h>
#include <tchar.h>
#include "CommRoutine.h"
#define TVTEST_PLUGIN_CLASS_IMPLEMENT
#include "TVTestPlugin.h"
#include "CCaptionPainter.h"

#define TVTEST_VIDEO_CONTAINER_CLASS_TEXT		_T("TVTest Video Container")

typedef struct _CH_CAPTION_PID {
	int		ch;					// Channel.
	int		serviceId;			// ServiceID.
	int		pid;				// The pid of caption(subtitle).
} CH_CAPTION_PID, *PCH_CAPTION_PID;

/*
 * The Mapping table of Caption PID.
 * TODO: Dynamically detect.
 */
CH_CAPTION_PID g_pidMap[] = {
	{786186481, 102, 0x230},	//BS: NHK2
	{786186482, 103, 0x130},	//BS: NHK Hi
	{783696081, 181, 0x114},	//BS: Fuji
	{771113009, 171, 0x145},	//BS: Japan
	{768557072, 151, 0x145},	//BS: Asahi
	{768557073, 161, 0x114}		//BS: BS-i
};

void GetCaptionPid(PBYTE pBuf, WORD ServiceID);

class CTVCaption : public TVTest::CTVTestPlugin
{
private:
	HWND m_hVideoContainer;				// The Video Container of TVTest
	int m_iChannel;						// Current Channel Value.
	WORD m_wServiceID;					// Current ServiceID.
	WORD m_wPMTPid;
	int m_iCaptionPid;					// Current Caption PID.
	BOOL m_bEnable;
	CCaptionPainter * m_pPainter;		// CaptionPainter.

	// EventCallBack
	static LRESULT CALLBACK EventCallback(UINT Event,LPARAM lParam1,LPARAM lParam2,void *pClientData);

	// StreamCallBack
	static BOOL CALLBACK StreamCallback(BYTE *pData,void *pClientData);

	// The proc of EnumChildWindows()
	static BOOL CALLBACK CTVCaption::EnumWindowProc(HWND hwnd, LPARAM lParam);

	// The proc of thread that try to get the channel infomation.
	static unsigned __stdcall CTVCaption::ReGetChannelInfo(LPVOID lpParameter);

	// internal method. Finds the video container.
	bool FindVideoContainer();

	// internal method. Shows video container's size. (FOR DEBUG)
	void ShowVideoSize();

	WORD GetCaptionPID(PBYTE pBuf, WORD ServiceID);

public:

	CTVCaption()
	{
		m_hVideoContainer = NULL;
		m_pPainter = NULL;
		m_bEnable = FALSE;
	}

	~CTVCaption(){};

	virtual bool GetPluginInfo(TVTest::PluginInfo *pInfo);
	virtual bool Initialize();
	virtual bool Finalize();

	// Sets the Video Container Handle
	void SetVideoHWND(HWND hWnd);

	// Gets the Video Container Handle
	HWND GetVideoHWND() { return this->m_hVideoContainer;}
	
	// Sets current channel information.
	void SetChannelInfo(int iChannel, WORD wServiceID);

	// Gets caption pid.
	int GetCaptionPID();

	bool GetServiceInfo(TVTest::ServiceInfo *pInfo) {
		int index = this->m_pApp->GetService();
		return this->m_pApp->GetServiceInfo(index, pInfo);
	}

	bool IsEnable() { return m_bEnable;}
	void SetEnable(bool bEnable) { m_bEnable = bEnable;}
};

bool CTVCaption::GetPluginInfo(TVTest::PluginInfo *pInfo)
{
	pInfo->Type           = TVTest::PLUGIN_TYPE_NORMAL;
	pInfo->Flags          = 0;
	pInfo->pszPluginName  = L"TV Caption";
	pInfo->pszCopyright   = L"Public Domain";
	pInfo->pszDescription = L"Show Caption Of BS";
	return true;
}

bool CTVCaption::Finalize()
{
	if(this->m_pPainter) {
		delete this->m_pPainter;
		this->m_pPainter = NULL;
	}

	return true;
}

bool CTVCaption::Initialize()
{
	DbgString(DBG_PREFIX _T("Initialize Start...\n"));

	this->FindVideoContainer();

	m_pApp->SetEventCallback(EventCallback,this);
	m_pApp->SetStreamCallback(0,StreamCallback,this);

	return true;
}

void CTVCaption::SetVideoHWND(HWND hWnd)
{
	if(hWnd == this->m_hVideoContainer)
		return;

	this->m_hVideoContainer = hWnd;

	DbgString(DBG_PREFIX _T("VideoHWND: %x\n"), hWnd);

	// The video container handle has been changed, we must
	// recreate CCaptionPainter.

	if(this->m_pPainter)
		delete this->m_pPainter;

	this->m_pPainter = new CCaptionPainter(this->m_hVideoContainer);
	if(!this->m_pPainter->Initialize()) {
		MessageBox(NULL, _T("CaptionDLL Initialize Failed"), _T("Warning"), MB_OK);
		delete this->m_pPainter;
		this->m_pPainter = NULL;
	}
}

BOOL CALLBACK CTVCaption::EnumWindowProc(HWND hwnd, LPARAM lParam)
{
	CTVCaption *pThis = static_cast<CTVCaption*>((void *)lParam);

	if(!pThis)
		return FALSE;

	HWND target = FindWindowEx(hwnd, NULL, TVTEST_VIDEO_CONTAINER_CLASS_TEXT, NULL);

	if(!target) {
		EnumChildWindows(hwnd, CTVCaption::EnumWindowProc, (LPARAM)(void *)pThis);

		if(pThis->GetVideoHWND())
			return FALSE;
	}
	else {
		pThis->SetVideoHWND(target);
		return FALSE;
	}

	return TRUE;
}

bool CTVCaption::FindVideoContainer()
{
	if(this->m_hVideoContainer)
		return TRUE;

	HWND hAp = this->m_pApp->GetAppWindow();

	EnumChildWindows(hAp, CTVCaption::EnumWindowProc, (LPARAM)(void *)this);

	return this->m_hVideoContainer != NULL;
}

void CTVCaption::ShowVideoSize()
{
	RECT rect;
	if(this->m_hVideoContainer) {

		GetWindowRect(this->m_hVideoContainer, &rect);
		int videoWidth = rect.right - rect.left;
		int videoHight = rect.bottom - rect.top;

		DbgString(DBG_PREFIX _T("PosX %d, PosY %d, AP Size: W %d H %d"), rect.left, rect.top, videoWidth, videoHight);
	}
	else {
		DbgString(DBG_PREFIX _T("video container handle = null\n"));
	}
}

void CTVCaption::SetChannelInfo(int iChannel, WORD wServiceID)
{
	 m_iChannel = iChannel;
	 m_wServiceID = wServiceID;

	this->m_iCaptionPid = 0;

	DbgString(DBG_PREFIX _T("Channel: %d, SID: %d, CaptionPID: %d\n"), iChannel, wServiceID, this->m_iCaptionPid);
}

int CTVCaption::GetCaptionPID()
{
	return m_iCaptionPid;
}

unsigned __stdcall CTVCaption::ReGetChannelInfo(LPVOID lpParameter)
{
	CTVCaption *pThis = static_cast<CTVCaption*>(lpParameter);

	TVTest::ServiceInfo Info;

	// Until success.
	while(!pThis->GetServiceInfo(&Info)) {
		Sleep(1000);
	}

	pThis->m_pPainter->Clear();

	// Only use ServiceID in BS.
	pThis->SetChannelInfo(0, Info.ServiceID);

	return 0;
}

LRESULT CALLBACK CTVCaption::EventCallback(UINT Event, LPARAM lParam1, LPARAM lParam2, void *pClientData)
{
	CTVCaption *pThis = static_cast<CTVCaption*>(pClientData);

	switch (Event) {
		case TVTest::EVENT_CHANNELCHANGE:
		case TVTest::EVENT_SERVICECHANGE:
			{
				// Create thread to get new channel information.
				_beginthreadex(
					NULL,
					0,
					CTVCaption::ReGetChannelInfo,
					pThis,
					0,
					NULL
					);
			}
			break;

		case TVTest::EVENT_PLUGINENABLE:
			{
				pThis->SetEnable((BOOL)lParam1);
				if(!lParam1) {
					// Channel has been changed. Clear all captions.
					pThis->m_pPainter->Clear();
				}

				DbgString(DBG_PREFIX _T("Plugin Enable: %d\n"), lParam1);

				return TRUE;
			}
			break;
	}

	return TRUE;
}

BOOL CALLBACK CTVCaption::StreamCallback(BYTE *pData,void *pClientData)
{
	CTVCaption *pThis = static_cast<CTVCaption*>(pClientData);

	if(!pThis)
		return TRUE;

	if(!pThis->IsEnable())
		return TRUE;

	int pid = pThis->GetCaptionPID(pData, pThis->m_wServiceID);

	if(pid == 0)
		return TRUE;

	// Parses packet.
	PBYTE pbPacket = pData;
	unsigned char ucSync = pbPacket[0];
	unsigned char ucTsErr = (pbPacket[1]&0x80)>>7;
	unsigned char ucPayloadStartFlag = (pbPacket[1]&0x40)>>6;
	unsigned char ucPriority = (pbPacket[1]&0x20)>>5;
	unsigned short usPID = ((unsigned short)(pbPacket[1]&0x1F))<<8 | pbPacket[2];
	unsigned char ucScramble = (pbPacket[3]&0xC0)>>6;
	unsigned char ucAdaptFlag = (pbPacket[3]&0x20)>>5;
	unsigned char ucPayloadFlag = (pbPacket[3]&0x10)>>4;
	unsigned char ucCounter = (pbPacket[3]&0x0F);

	if(usPID != pid)
		return TRUE;

	if(pThis->m_pPainter)
		pThis->m_pPainter->AddCaptionPacket(pbPacket);

	return TRUE;
}

TVTest::CTVTestPlugin *CreatePluginClass()
{
	return new CTVCaption();
}

/*
 * 2008-12-14
 * Parse PAT & PMT, try to find out caption pid.
 */
WORD CTVCaption::GetCaptionPID(PBYTE pBuf, WORD ServiceID)
{
	PBYTE pbPacket = pBuf;
	unsigned char ucSync = pbPacket[0];
	unsigned char ucTsErr = (pbPacket[1]&0x80)>>7;
	unsigned char ucPayloadStartFlag = (pbPacket[1]&0x40)>>6;
	unsigned char ucPriority = (pbPacket[1]&0x20)>>5;
	unsigned short usPID = ((unsigned short)(pbPacket[1]&0x1F))<<8 | pbPacket[2];
	unsigned char ucScramble = (pbPacket[3]&0xC0)>>6;
	unsigned char ucAdaptFlag = (pbPacket[3]&0x20)>>5;
	unsigned char ucPayloadFlag = (pbPacket[3]&0x10)>>4;
	unsigned char ucCounter = (pbPacket[3]&0x0F);

	if(usPID == 0x00) {
		// PAT

		PBYTE pData = pBuf + 13;

		do {

			// Copy From BonDriver
			WORD wProgramID	= ((WORD)pData[0] << 8) | (WORD)pData[1];				// +1,2
			WORD wPID		= ((WORD)(pData[2] & 0x1FU) << 8) | (WORD)pData[3];		// +3,4

			if(wProgramID == 0xFFFF)
				return this->m_iCaptionPid;

			if(wProgramID == this->m_wServiceID) {
				this->m_wPMTPid = wPID;
			}

			pData += 4;

		}while(pData < pBuf + 188);

		return this->m_iCaptionPid;
	}


	if(usPID == this->m_wPMTPid) {

		PBYTE pData = pBuf + 15;
		BYTE high = pData[0] % 0x10;
		pData++;
        BYTE low = pData[0];
        INT length = high * 0x100 + low;
		pData += length + 1;

		do {

			// Copy From BonDriver
			BYTE byStreamTypeID = pData[0];													// +0
			WORD wEsPID = ((WORD)(pData[1] & 0x1FU) << 8) | (WORD)pData[2];					// +1,2	
			WORD wDescLen = ((WORD)(pData[3] & 0x0FU) << 8) | (WORD)pData[4];				// +3,4

			if(byStreamTypeID == 0x6) {
				this->m_iCaptionPid = wEsPID;
				DbgString(DBG_PREFIX _T("*Caption PID: %x\n"), this->m_iCaptionPid);
			}

			pData += (wDescLen + 5);

		} while(pData < pBuf + 184);
	}

	return this->m_iCaptionPid;
}